//-------------------------------------------------------------------------------------------------
// IntVector2.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_IntVector2_
#define _include_IntVector2_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.hpp"
#include "engineMath.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		IntVector2
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class IntVector2
{
public:
	int x; // these are an exception to the "no-public-members" rule of thumb
	int y; // ...for the sake of clarity and brevity in use throughout the code

	static const IntVector2 ZERO;
	static const IntVector2 ONE;

	/// Construction / Destruction
	~IntVector2(); // destructor
	IntVector2(); // default constructor
	IntVector2( const IntVector2& vec ); // copy constructor
	IntVector2( const int x, const int y ); // construct from x,y components

	/// Methods
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const float SetLength( float newLength );
	void SetXY( const int new_x, const int new_y );
	void GetXY( int& get_x, int& get_y ) const;
	IntVector2& Clamp( const IntVector2& min, const IntVector2& max );
	IntVector2& SetFromCosSinRadians( double radians );
	IntVector2& SetFromCosSinDegrees( double degrees );
	IntVector2& SetFromRadiusCosSinRadians( double radius, double radians );
	IntVector2& SetFromRadiusCosSinDegrees( double radius, double degrees );
	double GetDegrees() const;
	double GetRadians() const;
	const int CalcMaxComponent() const;
	const int CalcMinComponent() const;
	const IntVector2 Lerp( const IntVector2& rhs, const float rhs_fraction ) const;
	const bool IsZero() const;
	const bool IsOne() const;

	/// Operators
	const IntVector2& operator = ( const IntVector2& rhs );
	const IntVector2& operator += ( const IntVector2& rhs );
	const IntVector2& operator -= ( const IntVector2& rhs );
	const IntVector2& operator *= ( const int scale );
	const IntVector2& operator *= ( const float scale );
	const IntVector2& operator /= ( const int divisor );
	const IntVector2& operator /= ( const float divisor );

	const IntVector2 operator - () const;
	const IntVector2 operator - ( const IntVector2& rhs ) const;
	const IntVector2 operator + ( const IntVector2& rhs ) const;
	const IntVector2 operator * ( const int scale ) const;
	const IntVector2 operator * ( const float scale ) const;
	const int operator | ( const IntVector2& rhs ) const; // dot product
	const int operator % ( const IntVector2& rhs ) const; // 2d cross product (scalar)

	const int operator [] ( const int index ) const;
	int& operator [] ( const int index );

	const bool operator == ( const IntVector2& rhs ) const;
	const bool operator != ( const IntVector2& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const IntVector2 operator * ( const int scale, const IntVector2& rhs );
const IntVector2 operator * ( const float scale, const IntVector2& rhs );
const bool EpsilonEqual( const IntVector2& lhs, const IntVector2& rhs, const int epsilon );


//-------------------------------------------------------------------------------------------------
// IntVector2::~IntVector2()
//-------------------------------------------------------------------------------------------------
inline IntVector2::~IntVector2()
{
}


//-------------------------------------------------------------------------------------------------
// IntVector2::IntVector2() - default constructor
//-------------------------------------------------------------------------------------------------
inline IntVector2::IntVector2()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// IntVector2::IntVector2( const IntVector2& vec ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline IntVector2::IntVector2( const IntVector2& vec )
:
x( vec.x ),
y( vec.y )
{
}


//-------------------------------------------------------------------------------------------------
// IntVector2( int, int )
//-------------------------------------------------------------------------------------------------
inline IntVector2::IntVector2( const int x, const int y )
:
x( x ),
y( y )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator = ( const IntVector2& rhs )
{
	if( this == &rhs )
		return *this;

	x = rhs.x;
	y = rhs.y;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator += ( const IntVector2& rhs )
{
	x += rhs.x;
	y += rhs.y;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator -= ( const IntVector2& rhs )
{
	x -= rhs.x;
	y -= rhs.y;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator *= ( const int scale )
{
	x *= scale;
	y *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator *= ( const float scale )
{
	x = (int)( (float)x * scale );
	y = (int)( (float)y * scale );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator /= ( const int divisor )
{
	if( divisor )
	{
		x /= divisor;
		y /= divisor;
	}
	else
	{
		x = 0;
		y = 0;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const IntVector2& IntVector2::operator /= ( const float divisor )
{
	if( divisor )
	{
		float invDivisor = 1.0f / divisor;
		x = (int)( (float)x * invDivisor );
		y = (int)( (float)y * invDivisor );
	}
	else
	{
		x = 0;
		y = 0;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const IntVector2 IntVector2::operator - () const
{
	return IntVector2( -x, -y );
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const IntVector2 IntVector2::operator - ( const IntVector2& rhs ) const
{
	return( IntVector2( x - rhs.x, y - rhs.y ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const IntVector2 IntVector2::operator + ( const IntVector2& rhs ) const
{
	return( IntVector2( x + rhs.x, y + rhs.y ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const IntVector2 IntVector2::operator * ( const int scale ) const
{
	return( IntVector2( x * scale, y * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const IntVector2 IntVector2::operator * ( const float scale ) const
{
	return( IntVector2(
			(int)( (float)x * scale ),
			(int)( (float)y * scale ) ) );
}


//-------------------------------------------------------------------------------------------------
// operator | (dot product)
//-------------------------------------------------------------------------------------------------
inline const int IntVector2::operator | ( const IntVector2& rhs ) const
{
	return( (x * rhs.x) + (y * rhs.y) );
}


//-------------------------------------------------------------------------------------------------
// operator % (2d scalar cross product)
//-------------------------------------------------------------------------------------------------
inline const int IntVector2::operator % ( const IntVector2& rhs ) const
{
	return( (x * rhs.y) - (y * rhs.x) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline const int IntVector2::operator [] ( const int index ) const
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 2 ), "IntVector2[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline int& IntVector2::operator [] ( const int index )
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 2 ), "IntVector2[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool IntVector2::operator == ( const IntVector2& rhs ) const
{
	return( (x == rhs.x) && (y == rhs.y) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool IntVector2::operator != ( const IntVector2& rhs ) const
{
	return( (x != rhs.x) || (y != rhs.y) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float IntVector2::CalcLengthSquared() const
{
	return (float)( x*x + y*y );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float IntVector2::CalcLength() const
{
	return( (float) sqrt( (double) (x*x + y*y) ) );
}


//-------------------------------------------------------------------------------------------------
// SetLength
//-------------------------------------------------------------------------------------------------
inline const float IntVector2::SetLength( float newLength )
{
	float oldLength = CalcLength();
	if( oldLength )
	{
		float scale = newLength / oldLength;
		(*this) *= scale;
	}

	return oldLength;
}


//-------------------------------------------------------------------------------------------------
// SetXY
//-------------------------------------------------------------------------------------------------
inline void IntVector2::SetXY( const int new_x, const int new_y )
{
	x = new_x;
	y = new_y;
}


//-------------------------------------------------------------------------------------------------
// GetXY
//-------------------------------------------------------------------------------------------------
inline void IntVector2::GetXY( int& get_x, int& get_y ) const
{
	get_x = x;
	get_y = y;
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline IntVector2& IntVector2::Clamp( const IntVector2& min, const IntVector2& max )
{
	/// Clamp X component
	if( x > max.x )
		x = max.x;
	else if( x < min.x )
		x = min.x;

	/// Clamp Y component
	if( y > max.y )
		y = max.y;
	else if( y < min.y )
		y = min.y;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromCosSinRadians
//-------------------------------------------------------------------------------------------------
inline IntVector2& IntVector2::SetFromCosSinRadians( double radians )
{
	x = static_cast<int>( cos( radians ) );
	y = static_cast<int>( sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromCosSinDegrees
//-------------------------------------------------------------------------------------------------
inline IntVector2& IntVector2::SetFromCosSinDegrees( double degrees )
{
	double radians = DEG2RAD( degrees );
	x = static_cast<int>( cos( radians ) );
	y = static_cast<int>( sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromRadiusCosSinRadians
//-------------------------------------------------------------------------------------------------
inline IntVector2& IntVector2::SetFromRadiusCosSinRadians( double radius, double radians )
{
	x = static_cast<int>( radius * cos( radians ) );
	y = static_cast<int>( radius * sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// SetFromRadiusCosSinDegrees
//-------------------------------------------------------------------------------------------------
inline IntVector2& IntVector2::SetFromRadiusCosSinDegrees( double radius, double degrees )
{
	double radians = DEG2RAD( degrees );
	x = static_cast<int>( radius * cos( radians ) );
	y = static_cast<int>( radius * sin( radians ) );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// GetDegrees
//-------------------------------------------------------------------------------------------------
inline double IntVector2::GetDegrees() const
{
	return RAD2DEG( atan2( y, x ) );
}


//-------------------------------------------------------------------------------------------------
// GetRadians
//-------------------------------------------------------------------------------------------------
inline double IntVector2::GetRadians() const
{
	return atan2( y, x );
}


//-------------------------------------------------------------------------------------------------
// CalcMaxComponent
//-------------------------------------------------------------------------------------------------
inline const int IntVector2::CalcMaxComponent() const
{
	/// Assume x is the largest
	int max = x;

	/// Check if y is bigger
	if( y > max )
		max = y;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// CalcMinComponent
//-------------------------------------------------------------------------------------------------
inline const int IntVector2::CalcMinComponent() const
{
	/// Assume x is the smallest
	int min = x;

	/// Check if y is smaller
	if( y < min )
		min = y;

	return( min );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const IntVector2 IntVector2::Lerp( const IntVector2& rhs, const float rhs_fraction ) const
{
	return IntVector2(
		x + (int)( rhs_fraction * (float)(rhs.x - x) ),
		y + (int)( rhs_fraction * (float)(rhs.y - y) ));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, IntVector2)
//-------------------------------------------------------------------------------------------------
inline const IntVector2 operator * ( const int scale, const IntVector2& rhs )
{
	return( IntVector2( scale * rhs.x, scale * rhs.y ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const IntVector2& lhs, const IntVector2& rhs, const int epsilon )
{
	return( fabs( lhs.x - rhs.x ) < epsilon &&
			fabs( lhs.y - rhs.y ) < epsilon );
}


//-------------------------------------------------------------------------------------------------
// IsZero
//-------------------------------------------------------------------------------------------------
inline const bool IntVector2::IsZero() const
{
	if( (x != 0) || (y != 0) )
		return false;
	else
		return true;
}


//-------------------------------------------------------------------------------------------------
// IsOne
//-------------------------------------------------------------------------------------------------
inline const bool IntVector2::IsOne() const
{
	if( (x != 1) || (y != 1) )
		return false;
	else
		return true;
}


#endif // _include_IntVector2_
