//-------------------------------------------------------------------------------------------------
// IntVector3.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_IntVector3_
#define _include_IntVector3_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "IntVector2.hpp"
#include "engineCommon.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		IntVector3
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class IntVector3
{
public:
	int x; // these are an exception to the "no-public-members" rule of thumb
	int y; // ...for the sake of clarity and brevity in use throughout the code
	int z;

	static const IntVector3 ZERO;
	static const IntVector3 ONE;

	/// Construction / Destruction
	~IntVector3(); // destructor
	IntVector3(); // default constructor
	IntVector3( const IntVector3& vec ); // copy constructor
	IntVector3( const IntVector2& vec ); // construct from 2d vector
	IntVector3( const int x, const int y, const int z );

	/// Interface methods
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const float CalcLengthXY() const;
	const float CalcLengthYZ() const;
	const float CalcLengthXZ() const;
	const float SetLength( float newLength );

	void SetXYZ( const int new_x, const int new_y, const int new_z );
	void GetXYZ( int& get_x, int& get_y, int& get_z ) const;
	const IntVector3& Clamp( const IntVector3& min, const IntVector3& max );
	const int CalcMaxComponent() const;
	const int CalcMinComponent() const;
	const IntVector3 Lerp( const IntVector3& rhs, const float rhs_fraction ) const;

	/// Operators
	const IntVector3& operator = ( const IntVector3& rhs );
	const IntVector3& operator += ( const IntVector3& rhs );
	const IntVector3& operator -= ( const IntVector3& rhs );
	const IntVector3& operator *= ( const int scale );
	const IntVector3& operator *= ( const float scale );
	const IntVector3& operator /= ( const int divisor );
	const IntVector3& operator /= ( const float divisor );

	const IntVector3 operator - () const;
	const IntVector3 operator - ( const IntVector3& rhs ) const;
	const IntVector3 operator + ( const IntVector3& rhs ) const;
	const IntVector3 operator * ( const int scale ) const;
	const IntVector3 operator * ( const float scale ) const;
	const IntVector3 operator % ( const IntVector3& rhs ) const; // cross product
	const int operator | ( const IntVector3& rhs ) const; // dot product

	const int operator [] ( const int index ) const;
	int& operator [] ( const int index );

	const bool operator == ( const IntVector3& rhs ) const;
	const bool operator != ( const IntVector3& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const IntVector3 operator * ( const int scale, const IntVector3& rhs );
const IntVector3 operator * ( const float scale, const IntVector3& rhs );
const bool EpsilonEqual( const IntVector3& lhs, const IntVector3& rhs, const int epsilon );


//-------------------------------------------------------------------------------------------------
// IntVector3::~IntVector3()
//-------------------------------------------------------------------------------------------------
inline IntVector3::~IntVector3()
{
}


//-------------------------------------------------------------------------------------------------
// IntVector3::IntVector3() - default constructor
//-------------------------------------------------------------------------------------------------
inline IntVector3::IntVector3()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// IntVector3::IntVector3( const IntVector3& vec ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline IntVector3::IntVector3( const IntVector3& vec )
:
x( vec.x ),
y( vec.y ),
z( vec.z )
{
}


//-------------------------------------------------------------------------------------------------
// IntVector3::IntVector3( const IntVector2& vec ) - construct from 2d vector
//-------------------------------------------------------------------------------------------------
inline IntVector3::IntVector3( const IntVector2& vec )
:
x( vec.x ),
y( vec.y ),
z( 0 )
{
}


//-------------------------------------------------------------------------------------------------
// IntVector3( int, int, int )
//-------------------------------------------------------------------------------------------------
inline IntVector3::IntVector3( const int x, const int y, const int z )
:
x( x ),
y( y ),
z( z )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator = ( const IntVector3& rhs )
{
	if( this == &rhs )
		return *this;

	x = rhs.x;
	y = rhs.y;
	z = rhs.z;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator += ( const IntVector3& rhs )
{
	x += rhs.x;
	y += rhs.y;
	z += rhs.z;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator -= ( const IntVector3& rhs )
{
	x -= rhs.x;
	y -= rhs.y;
	z -= rhs.z;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator *= ( const int scale )
{
	x *= scale;
	y *= scale;
	z *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator *= ( const float scale )
{
	x = (int)( (float)x * scale );
	y = (int)( (float)y * scale );
	z = (int)( (float)z * scale );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator /= ( const int divisor )
{
	if( divisor )
	{
		x /= divisor;
		y /= divisor;
		z /= divisor;
	}
	else
	{
		x = 0;
		y = 0;
		z = 0;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::operator /= ( const float divisor )
{
	if( divisor )
	{
		float invDivisor = 1.0f / divisor;
		x = (int)( (float)x * invDivisor );
		y = (int)( (float)y * invDivisor );
		z = (int)( (float)z * invDivisor );
	}
	else
	{
		x = 0;
		y = 0;
		z = 0;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::operator - () const
{
	return IntVector3( -x, -y, -z );
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::operator - ( const IntVector3& rhs ) const
{
	return( IntVector3( x - rhs.x, y - rhs.y, z - rhs.z ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::operator + ( const IntVector3& rhs ) const
{
	return( IntVector3( x + rhs.x, y + rhs.y, z + rhs.z ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::operator * ( const int scale ) const
{
	return( IntVector3( x * scale, y * scale, z * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::operator * ( const float scale ) const
{
	return( IntVector3(
			(int)( (float)x * scale ),
			(int)( (float)y * scale ),
			(int)( (float)z * scale ) ) );
}


//-------------------------------------------------------------------------------------------------
// operator % (cross product)
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::operator % ( const IntVector3& rhs ) const
{
	return IntVector3(	(y * rhs.z) - (z * rhs.y),
					(z * rhs.x) - (x * rhs.z),
					(x * rhs.y) - (y * rhs.x) );
}


//-------------------------------------------------------------------------------------------------
// operator | (dot product)
//-------------------------------------------------------------------------------------------------
inline const int IntVector3::operator | ( const IntVector3& rhs ) const
{
	return( (x * rhs.x) + (y * rhs.y) + (z * rhs.z) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline const int IntVector3::operator [] ( const int index ) const
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 3 ), "IntVector3[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline int& IntVector3::operator [] ( const int index )
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 3 ), "IntVector3[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool IntVector3::operator == ( const IntVector3& rhs ) const
{
	return( (x == rhs.x) && (y == rhs.y) && (z == rhs.z) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool IntVector3::operator != ( const IntVector3& rhs ) const
{
	return( (x != rhs.x) || (y != rhs.y) || (z != rhs.z) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float IntVector3::CalcLengthSquared() const
{
	return (float)( x*x + y*y + z*z );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float IntVector3::CalcLength() const
{
	return( (float) sqrt( (double)( x*x + y*y + z*z ) ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthXY
//-------------------------------------------------------------------------------------------------
inline const float IntVector3::CalcLengthXY() const
{
	return( (float) sqrt( (double)( x*x + y*y ) ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthYZ
//-------------------------------------------------------------------------------------------------
inline const float IntVector3::CalcLengthYZ() const
{
	return( (float) sqrt( (double)( y*y + z*z ) ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthXZ
//-------------------------------------------------------------------------------------------------
inline const float IntVector3::CalcLengthXZ() const
{
	return( (float) sqrt( (double)( x*x + z*z ) ) );
}


//-------------------------------------------------------------------------------------------------
// SetLength
//-------------------------------------------------------------------------------------------------
inline const float IntVector3::SetLength( float newLength )
{
	float oldLength = CalcLength();
	if( oldLength )
	{
		float scale = newLength / oldLength;
		(*this) *= scale;
	}

	return oldLength;
}


//-------------------------------------------------------------------------------------------------
// SetXYZ
//-------------------------------------------------------------------------------------------------
inline void IntVector3::SetXYZ( const int new_x, const int new_y, const int new_z )
{
	x = new_x;
	y = new_y;
	z = new_z;
}


//-------------------------------------------------------------------------------------------------
// GetXYZ
//-------------------------------------------------------------------------------------------------
inline void IntVector3::GetXYZ( int& get_x, int& get_y, int& get_z ) const
{
	get_x = x;
	get_y = y;
	get_z = z;
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline const IntVector3& IntVector3::Clamp( const IntVector3& min, const IntVector3& max )
{
	/// Clamp X component
	if( x > max.x )
		x = max.x;
	else if( x < min.x )
		x = min.x;

	/// Clamp Y component
	if( y > max.y )
		y = max.y;
	else if( y < min.y )
		y = min.y;

	/// Clamp Z component
	if( z > max.z )
		z = max.z;
	else if( z < min.z )
		z = min.z;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// CalcMaxComponent
//-------------------------------------------------------------------------------------------------
inline const int IntVector3::CalcMaxComponent() const
{
	/// Assume x is the largest
	int max = x;

	/// Check if y is bigger
	if( y > max )
		max = y;

	/// Check if z is bigger
	if( z > max )
		max = z;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// CalcMinComponent
//-------------------------------------------------------------------------------------------------
inline const int IntVector3::CalcMinComponent() const
{
	/// Assume x is the smallest
	int min = x;

	/// Check if y is smaller
	if( y < min )
		min = y;

	/// Check if z is smaller
	if( z < min )
		min = z;

	return( min );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const IntVector3 IntVector3::Lerp( const IntVector3& rhs, const float rhs_fraction ) const
{
	return IntVector3(
		x + (int)( rhs_fraction * (float)(rhs.x - x) ),
		y + (int)( rhs_fraction * (float)(rhs.y - y) ),
		z + (int)( rhs_fraction * (float)(rhs.z - z) ));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, IntVector3)
//-------------------------------------------------------------------------------------------------
inline const IntVector3 operator * ( const int scale, const IntVector3& rhs )
{
	return( IntVector3( scale * rhs.x, scale * rhs.y, scale * rhs.z ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const IntVector3& lhs, const IntVector3& rhs, const int epsilon )
{
	return( fabs( lhs.x - rhs.x ) < epsilon &&
			fabs( lhs.y - rhs.y ) < epsilon &&
			fabs( lhs.z - rhs.z ) < epsilon );
}


#endif // _include_IntVector3_
