//-------------------------------------------------------------------------------------------------
// IntVector4.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_IntVector4_
#define _include_IntVector4_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "IntVector3.hpp"
#include "engineCommon.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		IntVector4
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class IntVector4
{
public:
	int x; // these are an exception to the "no-public-members" rule of thumb
	int y; // ...for the sake of clarity and brevity in use throughout the code
	int z;
	int t;

	static const IntVector4 ZERO;
	static const IntVector4 ONE;

	/// Construction / Destruction
	~IntVector4(); // destructor
	IntVector4(); // default constructor
	IntVector4( const IntVector4& vec ); // copy constructor
	IntVector4( const IntVector3& vec ); // construct from 3d vector
	IntVector4( const IntVector2& vec ); // construct from 2d vector
	IntVector4( const int x, const int y, const int z, const int t );

	/// Interface methods
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const float CalcLengthXYZ() const;
	const float SetLength( float newLength );

	void SetXYZT( const int new_x, const int new_y, const int new_z, const int new_t );
	void GetXYZT( int& get_x, int& get_y, int& get_z, int& get_t ) const;
	const IntVector4& Clamp( const IntVector4& min, const IntVector4& max );
	const int CalcMaxComponent() const;
	const int CalcMinComponent() const;
	const IntVector4 Lerp( const IntVector4& rhs, const float rhs_fraction ) const;

	/// Operators
	const IntVector4& operator = ( const IntVector4& rhs );
	const IntVector4& operator += ( const IntVector4& rhs );
	const IntVector4& operator -= ( const IntVector4& rhs );
	const IntVector4& operator *= ( const int scale );
	const IntVector4& operator *= ( const float scale );
	const IntVector4& operator /= ( const int divisor );
	const IntVector4& operator /= ( const float divisor );

	const IntVector4 operator - () const;
	const IntVector4 operator - ( const IntVector4& rhs ) const;
	const IntVector4 operator + ( const IntVector4& rhs ) const;
	const IntVector4 operator * ( const int scale ) const;
	const IntVector4 operator * ( const float scale ) const;
	const int operator | ( const IntVector4& rhs ) const; // dot product

	const int operator [] ( const int index ) const;
	int& operator [] ( const int index );

	const bool operator == ( const IntVector4& rhs ) const;
	const bool operator != ( const IntVector4& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const IntVector4 operator * ( const int scale, const IntVector4& rhs );
const IntVector4 operator * ( const float scale, const IntVector4& rhs );
const bool EpsilonEqual( const IntVector4& lhs, const IntVector4& rhs, const int epsilon );


//-------------------------------------------------------------------------------------------------
// IntVector4::~IntVector4()
//-------------------------------------------------------------------------------------------------
inline IntVector4::~IntVector4()
{
}


//-------------------------------------------------------------------------------------------------
// IntVector4::IntVector4() - default constructor
//-------------------------------------------------------------------------------------------------
inline IntVector4::IntVector4()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// IntVector4::IntVector4( const IntVector4& vec ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline IntVector4::IntVector4( const IntVector4& vec )
:
x( vec.x ),
y( vec.y ),
z( vec.z ),
t( vec.t )
{
}


//-------------------------------------------------------------------------------------------------
// IntVector4::IntVector4( const IntVector3& vec ) - construct from 3d vector
//-------------------------------------------------------------------------------------------------
inline IntVector4::IntVector4( const IntVector3& vec )
:
x( vec.x ),
y( vec.y ),
z( vec.z ),
t( 0 )
{
}


//-------------------------------------------------------------------------------------------------
// IntVector4::IntVector4( const IntVector2& vec ) - construct from 2d vector
//-------------------------------------------------------------------------------------------------
inline IntVector4::IntVector4( const IntVector2& vec )
:
x( vec.x ),
y( vec.y ),
z( 0 ),
t( 0 )
{
}


//-------------------------------------------------------------------------------------------------
// IntVector4( int, int, int, int )
//-------------------------------------------------------------------------------------------------
inline IntVector4::IntVector4( const int x, const int y, const int z, const int t )
:
x( x ),
y( y ),
z( z ),
t( t )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator = ( const IntVector4& rhs )
{
	if( this == &rhs )
		return *this;

	x = rhs.x;
	y = rhs.y;
	z = rhs.z;
	t = rhs.t;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator += ( const IntVector4& rhs )
{
	x += rhs.x;
	y += rhs.y;
	z += rhs.z;
	t += rhs.t;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator -= ( const IntVector4& rhs )
{
	x -= rhs.x;
	y -= rhs.y;
	z -= rhs.z;
	t -= rhs.t;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator *= ( const int scale )
{
	x *= scale;
	y *= scale;
	z *= scale;
	t *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator *= ( const float scale )
{
	x = (int)( (float)x * scale );
	y = (int)( (float)y * scale );
	z = (int)( (float)z * scale );
	t = (int)( (float)t * scale );
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator /= ( const int divisor )
{
	if( divisor )
	{
		x /= divisor;
		y /= divisor;
		z /= divisor;
		t /= divisor;
	}
	else
	{
		x = 0;
		y = 0;
		z = 0;
		t = 0;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::operator /= ( const float divisor )
{
	if( divisor )
	{
		float invDivisor = 1.0f / divisor;
		x = (int)( (float)x * invDivisor );
		y = (int)( (float)y * invDivisor );
		z = (int)( (float)z * invDivisor );
		t = (int)( (float)t * invDivisor );
	}
	else
	{
		x = 0;
		y = 0;
		z = 0;
		t = 0;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const IntVector4 IntVector4::operator - () const
{
	return IntVector4( -x, -y, -z, -t );
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const IntVector4 IntVector4::operator - ( const IntVector4& rhs ) const
{
	return( IntVector4( x - rhs.x, y - rhs.y, z - rhs.z, y - rhs.t ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const IntVector4 IntVector4::operator + ( const IntVector4& rhs ) const
{
	return( IntVector4( x + rhs.x, y + rhs.y, z + rhs.z, t + rhs.t ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const IntVector4 IntVector4::operator * ( const int scale ) const
{
	return( IntVector4( x * scale, y * scale, z * scale, t * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const IntVector4 IntVector4::operator * ( const float scale ) const
{
	return( IntVector4(
			(int)( (float)x * scale ),
			(int)( (float)y * scale ),
			(int)( (float)z * scale ),
			(int)( (float)t * scale ) ) );
}


//-------------------------------------------------------------------------------------------------
// operator | (dot product)
//-------------------------------------------------------------------------------------------------
inline const int IntVector4::operator | ( const IntVector4& rhs ) const
{
	return( (x * rhs.x) + (y * rhs.y) + (z * rhs.z) + (t * rhs.t) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline const int IntVector4::operator [] ( const int index ) const
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 4 ), "IntVector4[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline int& IntVector4::operator [] ( const int index )
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 4 ), "IntVector4[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool IntVector4::operator == ( const IntVector4& rhs ) const
{
	return( (x == rhs.x) && (y == rhs.y) && (z == rhs.z) && (t == rhs.t) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool IntVector4::operator != ( const IntVector4& rhs ) const
{
	return( (x != rhs.x) || (y != rhs.y) || (z != rhs.z) || (t != rhs.t) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float IntVector4::CalcLengthSquared() const
{
	return (float)( x*x + y*y + z*z + t*t );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float IntVector4::CalcLength() const
{
	return( (float) sqrt( (double)( x*x + y*y + z*z + t*t ) ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthXYZ
//-------------------------------------------------------------------------------------------------
inline const float IntVector4::CalcLengthXYZ() const
{
	return( (float) sqrt( (double)( x*x + y*y + z*z ) ) );
}


//-------------------------------------------------------------------------------------------------
// SetLength
//-------------------------------------------------------------------------------------------------
inline const float IntVector4::SetLength( float newLength )
{
	float oldLength = CalcLength();
	if( oldLength )
	{
		float scale = newLength / oldLength;
		(*this) *= scale;
	}

	return oldLength;
}


//-------------------------------------------------------------------------------------------------
// SetXYZT
//-------------------------------------------------------------------------------------------------
inline void IntVector4::SetXYZT( const int new_x, const int new_y, const int new_z, const int new_t )
{
	x = new_x;
	y = new_y;
	z = new_z;
	t = new_t;
}


//-------------------------------------------------------------------------------------------------
// GetXYZT
//-------------------------------------------------------------------------------------------------
inline void IntVector4::GetXYZT( int& get_x, int& get_y, int& get_z, int& get_t ) const
{
	get_x = x;
	get_y = y;
	get_z = z;
	get_t = t;
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline const IntVector4& IntVector4::Clamp( const IntVector4& min, const IntVector4& max )
{
	/// Clamp X component
	if( x > max.x )
		x = max.x;
	else if( x < min.x )
		x = min.x;

	/// Clamp Y component
	if( y > max.y )
		y = max.y;
	else if( y < min.y )
		y = min.y;

	/// Clamp Z component
	if( z > max.z )
		z = max.z;
	else if( z < min.z )
		z = min.z;

	/// Clamp t component
	if( t > max.t )
		t = max.t;
	else if( t < min.t )
		t = min.t;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// CalcMaxComponent
//-------------------------------------------------------------------------------------------------
inline const int IntVector4::CalcMaxComponent() const
{
	/// Assume x is the largest
	int max = x;

	/// Check if y is bigger
	if( y > max )
		max = y;

	/// Check if z is bigger
	if( z > max )
		max = z;

	/// Check if t is bigger
	if( t > max )
		max = t;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// CalcMinComponent
//-------------------------------------------------------------------------------------------------
inline const int IntVector4::CalcMinComponent() const
{
	/// Assume x is the smallest
	int min = x;

	/// Check if y is smaller
	if( y < min )
		min = y;

	/// Check if z is smaller
	if( z < min )
		min = z;

	/// Check if t is smaller
	if( t < min )
		min = t;

	return( min );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const IntVector4 IntVector4::Lerp( const IntVector4& rhs, const float rhs_fraction ) const
{
	return IntVector4(
		x + (int)( rhs_fraction * (float)(rhs.x - x) ),
		y + (int)( rhs_fraction * (float)(rhs.y - y) ),
		z + (int)( rhs_fraction * (float)(rhs.z - z) ),
		t + (int)( rhs_fraction * (float)(rhs.t - t) ));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, IntVector4)
//-------------------------------------------------------------------------------------------------
inline const IntVector4 operator * ( const int scale, const IntVector4& rhs )
{
	return( IntVector4( scale * rhs.x, scale * rhs.y, scale * rhs.z, scale * rhs.t ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const IntVector4& lhs, const IntVector4& rhs, const int epsilon )
{
	return( fabs( lhs.x - rhs.x ) < epsilon &&
			fabs( lhs.y - rhs.y ) < epsilon &&
			fabs( lhs.z - rhs.z ) < epsilon &&
			fabs( lhs.t - rhs.t ) < epsilon );
}


#endif // _include_IntVector4_
