//-------------------------------------------------------------------------------------------------
// Vector3.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_Vector3_
#define _include_Vector3_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "Vector2.hpp"
#include "engineCommon.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		Vector3
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class Vector3
{
public:
	float x; // these are an exception to the "no-public-members" rule of thumb
	float y; // ...for the sake of clarity and brevity in use throughout the code
	float z;

	static const Vector3 ZERO;
	static const Vector3 ONE;

	/// Construction / Destruction
	~Vector3(); // destructor
	Vector3(); // default constructor
	Vector3( const Vector3& vec ); // copy constructor
	Vector3( const Vector2& vec ); // construct from 2d vector
	Vector3( const float x, const float y, const float z );

	/// Interface methods
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const float CalcLengthXY() const;
	const float CalcLengthYZ() const;
	const float CalcLengthXZ() const;
	const float Normalize();
	const float SetLength( float newLength );

	void SetXYZ( const float new_x, const float new_y, const float new_z );
	void GetXYZ( float& get_x, float& get_y, float& get_z ) const;
	const Vector3& Clamp( const Vector3& min, const Vector3& max );
	const float CalcMaxComponent() const;
	const float CalcMinComponent() const;
	const Vector3 Lerp( const Vector3& rhs, const float rhs_fraction ) const;

	/// Operators
	const Vector3& operator = ( const Vector3& rhs );
	const Vector3& operator += ( const Vector3& rhs );
	const Vector3& operator -= ( const Vector3& rhs );
	const Vector3& operator *= ( const float scale );
	const Vector3& operator /= ( const float divisor );

	const Vector3 operator - () const;
	const Vector3 operator - ( const Vector3& rhs ) const;
	const Vector3 operator + ( const Vector3& rhs ) const;
	const Vector3 operator * ( const float scale ) const;
	const Vector3 operator % ( const Vector3& rhs ) const; // cross product
	const float operator | ( const Vector3& rhs ) const; // dot product

	const float operator [] ( const int index ) const;
	float& operator [] ( const int index );

	const bool operator == ( const Vector3& rhs ) const;
	const bool operator != ( const Vector3& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const Vector3 operator * ( const float scale, const Vector3& rhs );
const bool EpsilonEqual( const Vector3& lhs, const Vector3& rhs, const float epsilon );


//-------------------------------------------------------------------------------------------------
// Vector3::~Vector3()
//-------------------------------------------------------------------------------------------------
inline Vector3::~Vector3()
{
}


//-------------------------------------------------------------------------------------------------
// Vector3::Vector3() - default constructor
//-------------------------------------------------------------------------------------------------
inline Vector3::Vector3()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// Vector3::Vector3( const Vector3& vec ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline Vector3::Vector3( const Vector3& vec )
:
x( vec.x ),
y( vec.y ),
z( vec.z )
{
}


//-------------------------------------------------------------------------------------------------
// Vector3::Vector3( const Vector2& vec ) - construct from 2d vector
//-------------------------------------------------------------------------------------------------
inline Vector3::Vector3( const Vector2& vec )
:
x( vec.x ),
y( vec.y ),
z( 0.0f )
{
}


//-------------------------------------------------------------------------------------------------
// Vector3( float, float, float )
//-------------------------------------------------------------------------------------------------
inline Vector3::Vector3( const float x, const float y, const float z )
:
x( x ),
y( y ),
z( z )
{
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const Vector3& Vector3::operator = ( const Vector3& rhs )
{
	if( this == &rhs )
		return *this;

	x = rhs.x;
	y = rhs.y;
	z = rhs.z;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const Vector3& Vector3::operator += ( const Vector3& rhs )
{
	x += rhs.x;
	y += rhs.y;
	z += rhs.z;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const Vector3& Vector3::operator -= ( const Vector3& rhs )
{
	x -= rhs.x;
	y -= rhs.y;
	z -= rhs.z;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const Vector3& Vector3::operator *= ( const float scale )
{
	x *= scale;
	y *= scale;
	z *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const Vector3& Vector3::operator /= ( const float divisor )
{
	if( divisor )
	{
		float invDivisor = 1.0f / divisor;
		x *= invDivisor;
		y *= invDivisor;
		z *= invDivisor;
	}
	else
	{
		x = 0.0f;
		y = 0.0f;
		z = 0.0f;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const Vector3 Vector3::operator - () const
{
	return Vector3( -x, -y, -z );
}


//-------------------------------------------------------------------------------------------------
// operator -
//-------------------------------------------------------------------------------------------------
inline const Vector3 Vector3::operator - ( const Vector3& rhs ) const
{
	return( Vector3( x - rhs.x, y - rhs.y, z - rhs.z ) );
}


//-------------------------------------------------------------------------------------------------
// operator +
//-------------------------------------------------------------------------------------------------
inline const Vector3 Vector3::operator + ( const Vector3& rhs ) const
{
	return( Vector3( x + rhs.x, y + rhs.y, z + rhs.z ) );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const Vector3 Vector3::operator * ( const float scale ) const
{
	return( Vector3( x * scale, y * scale, z * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator % (cross product)
//-------------------------------------------------------------------------------------------------
inline const Vector3 Vector3::operator % ( const Vector3& rhs ) const
{
	return Vector3(	(y * rhs.z) - (z * rhs.y),
					(z * rhs.x) - (x * rhs.z),
					(x * rhs.y) - (y * rhs.x) );
}


//-------------------------------------------------------------------------------------------------
// operator | (dot product)
//-------------------------------------------------------------------------------------------------
inline const float Vector3::operator | ( const Vector3& rhs ) const
{
	return( (x * rhs.x) + (y * rhs.y) + (z * rhs.z) );
}


//-------------------------------------------------------------------------------------------------
// operator [] const
//-------------------------------------------------------------------------------------------------
inline const float Vector3::operator [] ( const int index ) const
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 3 ), "Vector3[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator []
//-------------------------------------------------------------------------------------------------
inline float& Vector3::operator [] ( const int index )
{
//	ASSERT_MSG1( ( index >= 0 ) && ( index < 3 ), "Vector3[%d] was an out of range index!", index );
	return ( &x )[ index ];
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool Vector3::operator == ( const Vector3& rhs ) const
{
	return( (x == rhs.x) && (y == rhs.y) && (z == rhs.z) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool Vector3::operator != ( const Vector3& rhs ) const
{
	return( (x != rhs.x) || (y != rhs.y) || (z != rhs.z) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcLengthSquared() const
{
	return( x*x + y*y + z*z );
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcLength() const
{
	return( (float) sqrt( x*x + y*y + z*z ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthXY
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcLengthXY() const
{
	return( (float) sqrt( x*x + y*y ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthYZ
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcLengthYZ() const
{
	return( (float) sqrt( y*y + z*z ) );
}


//-------------------------------------------------------------------------------------------------
// CalcLengthXZ
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcLengthXZ() const
{
	return( (float) sqrt( x*x + z*z ) );
}


//-------------------------------------------------------------------------------------------------
// Normalize
//-------------------------------------------------------------------------------------------------
inline const float Vector3::Normalize()
{
	float length = CalcLength();
	if( length )
	{
		float invLength = 1.0f / length;
		x *= invLength;
		y *= invLength;
		z *= invLength;
	}

	return( length );
}


//-------------------------------------------------------------------------------------------------
// SetLength
//-------------------------------------------------------------------------------------------------
inline const float Vector3::SetLength( float newLength )
{
	float oldLength = Normalize();
	(*this) *= newLength;
	return oldLength;
}


//-------------------------------------------------------------------------------------------------
// SetXYZ
//-------------------------------------------------------------------------------------------------
inline void Vector3::SetXYZ( const float new_x, const float new_y, const float new_z )
{
	x = new_x;
	y = new_y;
	z = new_z;
}


//-------------------------------------------------------------------------------------------------
// GetXYZ
//-------------------------------------------------------------------------------------------------
inline void Vector3::GetXYZ( float& get_x, float& get_y, float& get_z ) const
{
	get_x = x;
	get_y = y;
	get_z = z;
}


//-------------------------------------------------------------------------------------------------
// Clamp
//-------------------------------------------------------------------------------------------------
inline const Vector3& Vector3::Clamp( const Vector3& min, const Vector3& max )
{
	/// Clamp X component
	if( x > max.x )
		x = max.x;
	else if( x < min.x )
		x = min.x;

	/// Clamp Y component
	if( y > max.y )
		y = max.y;
	else if( y < min.y )
		y = min.y;

	/// Clamp Z component
	if( z > max.z )
		z = max.z;
	else if( z < min.z )
		z = min.z;

	return *this;
}


//-------------------------------------------------------------------------------------------------
// CalcMaxComponent
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcMaxComponent() const
{
	/// Assume x is the largest
	float max = x;

	/// Check if y is bigger
	if( y > max )
		max = y;

	/// Check if z is bigger
	if( z > max )
		max = z;

	return( max );
}


//-------------------------------------------------------------------------------------------------
// CalcMinComponent
//-------------------------------------------------------------------------------------------------
inline const float Vector3::CalcMinComponent() const
{
	/// Assume x is the smallest
	float min = x;

	/// Check if y is smaller
	if( y < min )
		min = y;

	/// Check if z is smaller
	if( z < min )
		min = z;

	return( min );
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const Vector3 Vector3::Lerp( const Vector3& rhs, const float rhs_fraction ) const
{
	return Vector3( x + rhs_fraction * (rhs.x - x),
					y + rhs_fraction * (rhs.y - y),
					z + rhs_fraction * (rhs.z - z));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, Vector3)
//-------------------------------------------------------------------------------------------------
inline const Vector3 operator * ( const float scale, const Vector3& rhs )
{
	return( Vector3( scale * rhs.x, scale * rhs.y, scale * rhs.z ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const Vector3& lhs, const Vector3& rhs, const float epsilon )
{
	return( fabs( lhs.x - rhs.x ) < epsilon &&
			fabs( lhs.y - rhs.y ) < epsilon &&
			fabs( lhs.z - rhs.z ) < epsilon );
}


#endif // _include_Vector3_
